//----------------------------------------------------------------------------
// based on Borland WinSys Library
// Copyright (c) 1994, 1998 by Borland International, All Rights Reserved
//
// Modified by Mike Melnikov
//
// General Registry access & registration implementation
//   RegKey, RegValue, RegKeyIterator, RegValueIterator
//----------------------------------------------------------------------------

#ifndef _registry_h
#define _registry_h

typedef void*        PSECURITY_DESCRIPTOR;

#include "vector"

//----------------------------------------------------------------------------
// Direct registry access classes. Can read & iterate as well as create and
// write keys and value-data pairs
//

//
// class RegKey
// ~~~~~ ~~~~~~~
// Encapsulation of a registration key
//
class RegKeyIterator;
class RegKey
{
  public:
    typedef HKEY THandle;

    // Enumeration used to specify whether a key should be created (or
    // simply opened).
    //
    enum TCreateOK {
      CreateOK,             // Create key if it does not exist
      NoCreate              // Don't create, simply open
    };

    // Create or open a key given a base key and a subkeyname. Security
    // information is ignored in 16bit (and under Win95)
    // Can also provide an ok-to-create or open-only indicator.
    //
    RegKey(THandle baseKey,
            const _TCHAR far* keyName,
            REGSAM samDesired = KEY_ALL_ACCESS,
            TCreateOK createOK = CreateOK);

    // Construct a key give the current position of a regkey iterator
    //
    RegKey(const RegKeyIterator& iter, REGSAM samDesired = KEY_ALL_ACCESS);

    // Contruct a key that is an alias to an existing HKEY
    //
    RegKey(THandle aliasKey, bool shouldClose=false, const _TCHAR far* keyName = 0);
   ~RegKey();

    operator THandle() const;

    long DeleteKey(const _TCHAR far* subKeyName);
    long NukeKey(const _TCHAR far* subKeyName);
    long Flush() const;

    long GetSecurity(SECURITY_INFORMATION secInf, PSECURITY_DESCRIPTOR secDesc,
                     unsigned long* secDescSize);
    long SetSecurity(SECURITY_INFORMATION secInf, PSECURITY_DESCRIPTOR secDesc);

    // Saving & loading of this key & subKeys in NEW TEXT MODE
#ifdef _MY_REG_MODE
    void Write(const _TCHAR far* fileName);
    void Read(const _TCHAR far* fileName);
#endif

    // Saving & loading of this key & subKeys in WINDOWS TEXT MODE
    void WriteRegistry(const _TCHAR far* fileName);


    // Saving & loading of this key & subKeys
    //
    long Save(const _TCHAR far* fileName);
    long LoadKey(const _TCHAR far* subKeyName, const _TCHAR far* fileName);
    long ReplaceKey(const _TCHAR far* subKeyName, const _TCHAR far* newFileName,
                    const _TCHAR far* oldFileName);
    long Restore(const _TCHAR far* fileName, unsigned long options=0);
    long UnLoadKey(const _TCHAR far* subKeyName);

    //long NotifyChangeKeyValue();

    // Some of these are initialized at ctor
    //
    long QueryInfo(_TCHAR far* class_, unsigned long* classSize, unsigned long* subkeyCount,
                   unsigned long* maxSubkeySize, unsigned long* maxClassSize,
                   unsigned long* valueCount, unsigned long* maxValueName,
                   unsigned long* maxValueData, unsigned long* secDescSize,
                   FILETIME far* lastWriteTime);

    const _TCHAR far* GetName() const;
    unsigned long GetSubkeyCount() const;
    unsigned long GetValueCount() const;

    // Special predefined root keys
    //

  static RegKey ClassesRoot;
  static RegKey ClassesRootClsid;

  static RegKey& CurrentUser()     { return GetCurrentUser();    }
  static RegKey& LocalMachine()    { return GetLocalMachine();   }
  static RegKey& Users()           { return GetUsers();          }
  static RegKey& CurrentConfig()   { return GetCurrentConfig();  }
  static RegKey& DynData()         { return GetDynData();        }
  static RegKey& PerformanceData() { return GetPerformanceData();}

    // Accessors common/predefined registry keys
    //
    static RegKey& GetClassesRoot();
    static RegKey& GetClassesRootClsid();
    static RegKey& GetCurrentUser();
    static RegKey& GetLocalMachine();
    static RegKey& GetUsers();
    static RegKey& GetCurrentConfig();
    static RegKey& GetDynData();
    static RegKey& GetPerformanceData();

    // Friend iterators only...?
    //
    long EnumKey(int index, _TCHAR far* subKeyName, int subKeyNameSize) const;

    // Older, nameless subkey+(Default) value access
    //
    long SetDefValue(const _TCHAR far* subkeyName, unsigned long type,
                     const _TCHAR far* data, unsigned long dataSize);
    long QueryDefValue(const _TCHAR far* subkeyName, _TCHAR far* data,
                       unsigned long* dataSize) const;

    // Newer, named value set & get functions. Subkeys must be opened
    //
    long SetValue(const _TCHAR far* valName, unsigned long type, const unsigned char* data,
                  unsigned long dataSize) const;
    long SetValue(const _TCHAR far* valName, unsigned long data) const;
    long QueryValue(const _TCHAR far* valName, unsigned long* type, unsigned char* data,
                    unsigned long* dataSize) const;
    long DeleteValue(const _TCHAR far* valName) const;

    // Friend iterators only...?
    //
    long EnumValue(int index, _TCHAR far* valueName, unsigned long& valueNameSize,
                   unsigned long* type = 0, unsigned char* data = 0, unsigned long* dataSize = 0) const;

  protected:
    THandle   Key;          // This Key's Handle
    _TCHAR far* Name;         // This Key's Name
    unsigned long    SubkeyCount;  // Number of subkeys
    unsigned long    ValueCount;   // Number of value entries
    bool      ShouldClose;  // Should this key be closed on destruction

#ifdef _MY_REG_MODE
    void Write(FILE* fp,int lev);
    void Read(FILE* fp,int lev);
#endif
    void WriteRegistry(FILE* fp,CString parent);
};


//
// class RegValue
// ~~~~~ ~~~~~~~~~
// Encapsulation of a value-data entry within one registration key.
//
class RegValueIterator;
class RegValue {
  public:
    RegValue(const RegKey& key, const _TCHAR far* name);
    RegValue(const RegValueIterator& iter);
   ~RegValue();

    const _TCHAR far* GetName() const;
    const unsigned long GetDataType() const;
    const unsigned long GetDataSize() const;
    const unsigned char* GetData() const;
    operator unsigned long() const;
    operator const _TCHAR far*() const;

    // Set the data and possibly type of this value
    //
    long Set(unsigned long type, unsigned char* data, unsigned long dataSize);

    // VOVA LOOK HERE 
    long Set(long data);
    long Set(unsigned long data);
    long Set(const _TCHAR far* data);
    long Set(float data);
    long Set(int x,  int y);
    long Set(float x,float y);
    long Set(bool data);

    void Get(long& data) const;
    void Get(_TCHAR far* data) const;
    void Get(float& data) const;
    void Get(long& x, long& y) const;
    void Get(float& x,float& y) const;
    void Get(bool& data) const;

    void operator =(long v);
    void operator =(const _TCHAR far* v);
    void operator =(float v);
    void operator =(bool v);

    operator long() const;
    operator float() const;
    operator const bool() const;

    long Delete();

    void RetrieveOnDemand() const;

  private:
    const RegKey&  Key;      // Key that this value lives in
    const _TCHAR far* Name;     // Working name pointer
    _TCHAR far*       AName;    // Allocated space for name if needed

    // Data information. 
    unsigned long          DataType; // Type code for value data
    mutable unsigned char*  Data;     // Value data itself, allocated
    unsigned long          DataSize; // Size in bytes of Data
};

//
// class RegKeyIterator
// ~~~~~ ~~~~~~~~~~~~~~~
// Iterator for walking thru the subkeys of a key
//
class RegKeyIterator {
  public:
    RegKeyIterator(const RegKey& key);

    operator bool();

    unsigned long operator ++();
    unsigned long operator ++(int);
    unsigned long operator --();
    unsigned long operator --(int);
    unsigned long operator [](int index);
//    RegKey*      operator->() { return & operator[Current()];}


    int             Current() const;
    const RegKey&  BaseKey() const;

    void Reset();

  private:
    const RegKey& Key;
    int            Index;
};

//
// class RegValueIterator
// ~~~~~ ~~~~~~~~~~~~~~~~~
// Iterator for walking thru the values of a key
//
class RegValueIterator {
  public:
    RegValueIterator(const RegKey& regKey);

    operator bool();

    unsigned long operator ++();
    unsigned long operator ++(int);
    unsigned long operator --();
    unsigned long operator --(int);
    unsigned long operator [](int index);

    int            Current() const;
    const RegKey& BaseKey() const;

    void Reset();

  private:
    const RegKey& Key;
    int            Index;
};

//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// RegKey inlines
//

//
// Return the HANDLE identifying this registry key.
//
inline
RegKey::operator RegKey::THandle() const
{
  return Key;
}

//
// Enumerate the subkeys of this registry key
//
inline long
RegKey::EnumKey(int index, _TCHAR far* subKeyName, int subKeyNameSize) const
{
  return ::RegEnumKey(Key, index, subKeyName, subKeyNameSize);
  //::RegEnumKeyEx(); ??
}

//
// Delete the specified subkey of this registry key
//
inline long
RegKey::DeleteKey(const _TCHAR far* subKeyName)
{
  return ::RegDeleteKey(Key, subKeyName);
}

//
// Write the attribute of this key in the registry
// Note: 32 bit only
//
inline long
RegKey::Flush() const
{
  return ::RegFlushKey(Key);
}

//
// Retrieve a copy of the security descriptor protecting this registry key.
// Note: 32 bit only
//
inline long
RegKey::GetSecurity(SECURITY_INFORMATION secInf, PSECURITY_DESCRIPTOR secDesc,
                     unsigned long* secDescSize)
{
  return ::RegGetKeySecurity(Key, secInf, secDesc, secDescSize);
}

//
// Set the security descriptor of this key
// Note: 32 bit only
//
inline long
RegKey::SetSecurity(SECURITY_INFORMATION secInf, PSECURITY_DESCRIPTOR secDesc)
{
  return ::RegSetKeySecurity(Key, secInf, secDesc);
}

//
// Save this key and all of its subkeys and values to the specified file.
// Note: 32 bit only
//
inline long
RegKey::Save(const _TCHAR far* fileName)
{
  return ::RegSaveKey(Key, fileName, 0/*Security*/);
}

//
// Create a subkey under HKEY_USER or HKEY_LOCAL_MACHINE and stores
// registration information from a specified file into that subkey. This
// registration information is in the form of a hive. A hive is a discrete
// body of keys, subkeys, and values that is rooted at the top of the
// registry hierarchy. A hive is backed by a single file and .LOG file.
// Note: 32 bit only
//
inline long
RegKey::LoadKey(const _TCHAR far* subKeyName, const _TCHAR far* fileName)
{
  return ::RegLoadKey(Key, subKeyName, fileName);
}

//
// Replace the file backing this key and all its subkeys with another file,
// so that when the system is next started, the key and subkeys will have
// the values stored in the new file.
// Note: 32 bit only
//
inline long
RegKey::ReplaceKey(const _TCHAR far* subKeyName, const _TCHAR far* newFileName,
                    const _TCHAR far* oldFileName)
{
  return ::RegReplaceKey(Key, subKeyName, newFileName, oldFileName);
}

//
// Read the registry information in a specified file and copies it over this
// key. This registry information may be in the form of a key and multiple
// levels of subkeys.
// Note: 32 bit only
//
inline long
RegKey::Restore(const _TCHAR far* fileName, unsigned long options)
{
  return ::RegRestoreKey(Key, fileName, options);
}

//
// Unload this key and its subkeys from the registry.
//
// Note: 32 bit only
inline long
RegKey::UnLoadKey(const _TCHAR far* subKeyName)
{
  return ::RegUnLoadKey(Key, subKeyName);
}

//inline long RegKey::NotifyChangeKeyValue() {}

//
// Return the number of subkeys attached to this key
//
inline unsigned long
RegKey::GetSubkeyCount() const
{
  return SubkeyCount;
}

//
// Return the number of values attached to this key
//
inline unsigned long
RegKey::GetValueCount() const
{
  return ValueCount;
}

//
// Return a string identifying this key
//
inline const _TCHAR far*
RegKey::GetName() const
{
  return Name;
}

//
// Associate a value with this key
//
inline long
RegKey::SetValue(const _TCHAR far* valName, unsigned long type, const unsigned char* data,
                  unsigned long dataSize) const
{
  return ::RegSetValueEx(Key, valName, 0, type, data, dataSize);
}

//
// Associate a 4-byte value with this key
// Note: 32 bit only
//
inline long
RegKey::SetValue(const _TCHAR far* valName, unsigned long data) const
{
  return SetValue(valName, REG_DWORD, (unsigned char*)&data, sizeof data);
}

//
// Retrieve the value associated with the unnamed value for this key
//
inline long
RegKey::QueryValue(const _TCHAR far* valName, unsigned long* type, unsigned char* data,
                    unsigned long* dataSize) const
{
  return ::RegQueryValueEx(Key, valName, 0, type, data, dataSize);
}

//
// Remove a named value from this registry key
// Note: 32 bit only
//
inline long
RegKey::DeleteValue(const _TCHAR far* valName) const
{
  return ::RegDeleteValue(Key, valName);
}

//
// Set the default [unnamed] value associated with this key
//
inline long
RegKey::SetDefValue(const _TCHAR far* subkeyName, unsigned long type,
                     const _TCHAR far* data, unsigned long dataSize)
{
  return ::RegSetValue(Key, subkeyName, type, data, dataSize);
}

//
// Retrieve the default [unnamed] value associated with this key
//
inline long
RegKey::QueryDefValue(const _TCHAR far* subkeyName, _TCHAR far* data,
                       unsigned long* dataSize) const
{
  return ::RegQueryValue(Key, subkeyName, data, (long*)dataSize);
}

//
// Enumerate the values associated with this key.  Copy the value name and
// data block associated with the passed index.
// Note: 32 bit only
//
inline long
RegKey::EnumValue(int index, _TCHAR far* valueName, unsigned long& valueNameSize,
                   unsigned long* type, unsigned char* data, unsigned long* dataSize) const
{
  return ::RegEnumValue(Key, index, valueName, &valueNameSize, 0, type, data, dataSize);
}

//----------------------------------------------------------------------------
// RegValue inlines
//

//
// Return a string identifying this value
//
inline const _TCHAR far*
RegValue::GetName() const
{
  return Name;
}

//
// Return the type code for the data associated with this value
//
inline const unsigned long
RegValue::GetDataType() const
{
  return DataType;
}

//
// Return the size in bytes of the data associated with this value
//
inline const unsigned long
RegValue::GetDataSize() const
{
  return DataSize;
}

//
// Return the data associated with this value.
//
inline const unsigned char*
RegValue::GetData() const {
  RetrieveOnDemand();
  return Data;
}

inline void RegValue::Get(long& data) const 
  {sscanf((char*)GetData(),"%ld",&data);}
inline  void RegValue::Get(_TCHAR far* data) const 
  {data = (char*)GetData();}
inline  void RegValue::Get(float& data) const 
  {sscanf((char*)GetData(),"%f",&data);}
inline  void RegValue::Get(long& x,  long& y) const 
  {sscanf((char*)GetData(),"%ld %ld",&x,&y);}
inline  void RegValue::Get(float& x,float& y) const 
  {sscanf((char*)GetData(),"%f %f",&x,&y);}
inline  void RegValue::Get(bool& data) const 
  {long v; Get(v);data = v ? true: false;}

//
// Return the data associated with this value as a 32bit unsigned longeger
//
inline RegValue::operator unsigned long() const
{
  return *(unsigned long*)GetData();           // Assumes dataType==REG_DWORD
}

//
// Return the data associated with this value as a const char*
//
inline RegValue::operator const _TCHAR far*() const
{
  return (const _TCHAR far*)GetData();  // Assumes DataType==REG_SZ or sim.
}

//
// Set the data associated with this value
// Note: 32 bit only
//
inline long
RegValue::Set(unsigned long data)
{
  return Set(REG_DWORD, (unsigned char*)&data, sizeof data);
}

inline long
RegValue::Set(const _TCHAR far* data)
{
//  TString strData(data);
//  wchar_t* ustrData = strData;  
  return Set(REG_SZ, (unsigned char*)data, _tcslen(data));
}

inline long
RegValue::Set(long data)
{
  char str[100];sprintf(str,"%ld",data);
  return Set(str);
}

inline long
RegValue::Set(float data)
{
  char str[100];sprintf(str,"%f",data);
  return Set(str);
}

inline long
RegValue::Set(int x,  int y)
{
  char str[100];sprintf(str,"%d %d",x,y);
  return Set(str);
}

inline long
RegValue::Set(float x,float y)
{
  char str[100];sprintf(str,"%f %f",x,y);
  return Set(str);
}

inline long
RegValue::Set(bool data)
{
  return Set(data ? (long)1 : (long)0);
}



//
// Remove this value from its associated key.
// Note: the state of this value object becomes undefined.
//
inline long
RegValue::Delete()
{
  return Key.DeleteValue(Name);
}

//
// Set the data for this value to v
//
inline void RegValue::operator =(long v)    {Set(v);}
inline void RegValue::operator =(const _TCHAR far* v){Set(v);}
inline void RegValue::operator =(float v)            {Set(v);}
inline void RegValue::operator =(bool v)             {Set(v);}

inline RegValue::operator long() const
  {long v; Get(v); return v;}
inline RegValue::operator float() const
  {float  v; Get(v); return v;}
inline RegValue::operator const bool() const
  {bool v; Get(v); return v;}

//----------------------------------------------------------------------------
// RegKeyIterator inlines
//

//
// Create a subkey iterator for a registration key
//
inline
RegKeyIterator::RegKeyIterator(const RegKey& key)
:
  Key(key),
  Index(0)
{
}

//
// Test the validity of this iterator.  True if the iterator's
// index is greater than or equal to 0 and less than the number
// of subkeys.
//
inline
RegKeyIterator::operator bool()
{
  return Index >= 0 && Index < int(Key.GetSubkeyCount());
}

// Preincrement to the next subkey
inline unsigned long RegKeyIterator::operator ++(){return ++Index;}

// Postincrement to the next subkey
inline unsigned long RegKeyIterator::operator ++(int){return Index++;}

// Predecrement to the previous subkey
inline unsigned long RegKeyIterator::operator --(){return --Index;}

// Postdecrement to the previous subkey
inline unsigned long RegKeyIterator::operator --(int){return Index--;}

//
// Set the index of the iterator to the passed value.  Return the new index.
//
inline unsigned long
RegKeyIterator::operator [](int index)
{
  ASSERT((index >= 0) && (index < int(Key.GetSubkeyCount())));
  return Index = index;
}

// Return the index to the current subkey
inline int RegKeyIterator::Current() const { return Index;}

// Return the registration key this iterator is bound to
inline const RegKey& RegKeyIterator::BaseKey() const { return Key; }

// Reset the subkey index to zero
inline void RegKeyIterator::Reset() {Index = 0;}

//----------------------------------------------------------------------------
// RegValueIterator inlines
//

//
// Create a value iterator for a registration key
//
inline RegValueIterator::RegValueIterator(const RegKey& regKey):Key(regKey),Index(0){}

//
// Test the validity of this iterator.  True if the iterator's
// index is greater than or equal to 0 and less than the number
// of values.
//
inline
RegValueIterator::operator bool()
{
  return Index >= 0 && Index < int(Key.GetValueCount());
}

// Preincrement to the next value
inline unsigned long RegValueIterator::operator ++() {return ++Index;}

// Postincrement to the next value
inline unsigned long RegValueIterator::operator ++(int){return Index++;}

// Predecrement to the previous value
inline unsigned long RegValueIterator::operator --() { return --Index; }

// Postdecrement to the previous value
inline unsigned long RegValueIterator::operator --(int) { return Index--;}

// Set the index of the iterator to the passed value.  Return the new index.
inline unsigned long
RegValueIterator::operator [](int index)
{
  ASSERT((index >= 0) && (index < int(Key.GetValueCount())));
  return Index = index;
}

// Return the index to the current value
inline int RegValueIterator::Current() const { return Index; }

// Return the registration key this iterator is bound to
inline const RegKey& RegValueIterator::BaseKey() const {return Key;}

// Reset the value index to zero
inline void RegValueIterator::Reset() {Index = 0;}


inline TCHAR* strnewdup(LPCSTR str)
{
   ASSERT(str);
   return lstrcpy(new TCHAR[lstrlen(str)+1], str);
}

#endif

