// COPYRIGHT (C) 1993-1999 by Alexander Gavrilov.
// Modified by Mike Melnikov
#include "stdafx.h"

#include "RegularExpressions.h"
#include "Commands.h"

const int RTRUE  = -1,
          RFALSE =  0,
          ALMOST =  1;      // returned when a closure matches a NULL

const char ENDSTR   = '\0';
const char EOL      = '$';
const char BOL      = '^';
const char NEGATE   = '^';
const char CCL      = '[';
const char NCCL     = ']';
const char CCLEND   = ']';
const char ANY      = '.';
const char DASH     = '-';
const char OR       = '|';
const char ESCAPE   = '\\';
const char LPAREN   = '(';
const char RPAREN   = ')';
const char CLOSURE  = '*';
const char POS_CLO  = '+';
const char ZERO_ONE = '?';
const char LITCHAR  = 'c';
const char END_TERM = 'e';

static const char RegularCharacters[] =
{
  ZERO_ONE,
  CLOSURE,
  POS_CLO,
  ANY,
  CCL,
  CCLEND,
  NEGATE,
  EOL,
  ESCAPE,
  LPAREN,
  RPAREN,
  OR,
  ENDSTR
};

static const char *reExpression;         // offset for parsing

static bool		   ParseExpression(string&);
static bool        ParseTerm(string&);
static bool		   ParseFactor(string&);
static char        ParseEscape();
static int         ParseClosure(string&);
static void        ParseCCL(string&);
static const char *ParseDASH(string&, char);
static int          IsFactor();

static bool         Match( const char *s, const char *pattern, int& rstart, int& rlength );
static const char *SkipTerm(const char* pattern);
static int          MatchTerm(int, const char*, const char*);
static int          MatchOR(int, const char*, const char*);
static int          Match_0_1(int, const char*, const char*);
static int          MatchClosure(int, const char*, const char*, const char*);
static int          MatchCCL(char, const char*);



/* ===========================================================================
**                              RegularExpression
** =========================================================================*/

bool RegularExpression::CaseSensitive = true;
//--------------------------------------------------------------- constructors

RegularExpression::RegularExpression()
{
}

RegularExpression::RegularExpression( const string& s )
  : expression( s )
{
  MakePattern();
}

RegularExpression& RegularExpression::operator = ( const char* s )
{
  expression = s;

  if(CCommands::getValue("Regular",0)==0)
  {
      replace(expression,"*","[A-z0-9_.]*");
      replace(expression,"?",".");
  }

  if(!RegularExpression::CaseSensitive)
	  _tcsupr((char*)expression.c_str());

  MakePattern();
  ASSERT(expression[0]=='^');
  expression.erase(0,1);

  return *this;
}

bool RegularExpression::index( const char *s, int& rstart, int& rlength ) const
{
  return Match( s, rePattern.c_str(), rstart, rlength );
}

//----------------------------------------------------------------- comparison

bool RegularExpression::operator == ( const CString& s ) const
{
  int rstart, rlength;
  return index(s,rstart,rlength);
}


//----------------------------------------------------------------------------
//                         string functions
//----------------------------------------------------------------------------
void TrimLeft(CString& str,const char * lpszTargets)
{
#if _MSC_VER < 1200
	LPTSTR lpsz = str.LockBuffer();
	LPTSTR m_pchData = lpsz;

	while (*lpsz != '\0')
	{
		if (_tcschr(lpszTargets, *lpsz) == NULL)
			break;
		lpsz = _tcsinc(lpsz);
	}

	if (lpsz != m_pchData)
	{
		// fix up data and length
		int nDataLength = str.GetLength() - (lpsz - m_pchData);
		memmove(m_pchData, lpsz, (nDataLength+1)*sizeof(TCHAR));
		str.GetBufferSetLength(nDataLength);
	}
   str.UnlockBuffer();
#else
  str.TrimLeft(lpszTargets);
#endif
}

void TrimRight(CString& str,const char* lpszTargetList)
{
#if _MSC_VER < 1200
	LPTSTR lpsz = str.LockBuffer();
	LPTSTR m_pchData = lpsz;
	LPTSTR lpszLast = NULL;

	while (*lpsz != '\0')
	{
		if (_tcschr(lpszTargetList, *lpsz) != NULL)
		{
			if (lpszLast == NULL)
				lpszLast = lpsz;
		}
		else
			lpszLast = NULL;
		lpsz = _tcsinc(lpsz);
	}

	if (lpszLast != NULL)
	{
		// truncate at left-most matching character
		//*lpszLast = '\0';
		str.GetBufferSetLength(lpszLast - m_pchData);
	}
  str.UnlockBuffer();
#else
  str.TrimRight(lpszTargetList);
#endif
}

void replace(string& str,string::size_type start,string::size_type end, const char *new_str)
{
	str.erase(start,end-start);
    str.insert(start,new_str);
}

int replace(string& strThis,const RegularExpression& from, const char *to, bool all )
{
  int rstart, rlength, index = -1;
  size_t       to_len = to != 0 ? strlen(to) : 0;
  const char *str = strThis.c_str();
  int count = 0;
  while( from.index( str, rstart, rlength ) )
  {
    index = (str-strThis.c_str())+rstart;

    replace(strThis,index,index + rlength,to);
	
	count++;

    if( !all )
      break;
    else
      str += to_len;    // advance the beginning of search
  }
  return count;
}

int  replace(string& str,const string& from, const string& to, bool all)
{
  int count = 0;
  int index;
  int pos = 0;
  while( (index=str.find(from,pos))!=string::npos )
  {
    str.replace(index,from.size(),to);
	count++;
    if( !all )
      break;
    else
      pos = index + to.size();    // advance the beginning of search
  }
  return count;
}


//----------------------------------------------------------------------------
//                       internal functions
//----------------------------------------------------------------------------

// MakePattern() - "compile" the regular expression into rePattern

void RegularExpression::MakePattern()
{
  if(expression.size()==0)
  {
    rePattern.erase();
    return;
  }
  else
  {
    reExpression = expression.c_str();

    if( ParseExpression(rePattern) == false || *reExpression != ENDSTR )
    {
      rePattern.erase();
      return;
    }
    return;
  }
}


// ParseExpression() - Parse and translate an expression. Returns a pointer
// to the compiled expression, or NULL on error.

static bool ParseExpression(string& expr)
{
  string term;
  expr.erase();

  if(!ParseTerm(term))             // get the first term
    return false;

  while( *reExpression == OR )         // parse all subsequent terms
  {
    expr += OR;
    expr += term;
    expr += END_TERM;
    reExpression++;

    if(!ParseTerm(term))
      return false;
  }
  expr += term;
  expr += END_TERM;

  return true;
}

// ParseTerm() - parse and translate a term.  Returns a pointer to the
// compiled term or NULL on error.

static bool ParseTerm(string& term)
{
  string factor;
  term.erase();

  if( *reExpression == BOL )
  {
    term += reExpression[0];
    reExpression++;
  }
  do
  {
    if(!ParseFactor(factor))
      return false;

    term += factor;
  }
  while(IsFactor());			// parse all factors of this term

  return true;
}

// IsFactor() - returns RTRUE for a valid factor character


static int IsFactor()
{
  static const char nfac_chars[] = "^|)]+?*";

  return (strchr(nfac_chars, reExpression[0]) == NULL) ? RTRUE : RFALSE;
}

// ParseFactor() - parse and translate a factor.  Returns a pointer to the
// compiled factor or NULL on error.


static bool ParseFactor(string& factor)
{
  string tmp;
  factor.erase();

  switch( reExpression[0] )
  {
    case LPAREN:			// ( - parenthesised expression
      reExpression++;
      ParseExpression(tmp);
      factor += tmp;
      if( *reExpression != RPAREN )
	return(NULL);
      reExpression++;
      break;

    case CCL:				// [ - character class; Ex: [0-9]
      reExpression++;
      ParseCCL(tmp);
      factor += tmp;
      if( *reExpression != CCLEND )
        return(NULL);
      reExpression++;
      break;

    case ANY:				// .
    case EOL:				// $
      factor += reExpression[0];
      reExpression++;
      break;

    case ESCAPE :                       // backslash
      reExpression++;
      factor += LITCHAR;
      factor += ParseEscape();
      reExpression++;
      break;

    case CLOSURE:			// *
    case POS_CLO:			// +
    case ZERO_ONE:			// ?
    case NEGATE:			// ^
    case CCLEND:			// ]
    case RPAREN:			// )
    case OR:				// |
      return(NULL);			// not valid characters

    default:				// literal character
      factor += LITCHAR;
      factor += reExpression[0];
      reExpression++;
      break;
  }

  if( *reExpression == CLOSURE  ||        // check for closure
      *reExpression == ZERO_ONE ||
      *reExpression == POS_CLO )
  {
    if(ParseClosure(factor) == RFALSE)
      return false;
  }
  return true;
}

// ParseEscape() - returns ASCII value of character(s) following ESCAPE

static char ParseEscape()
{
  switch( reExpression[0] )
  {
    case 'b': reExpression++; return '\b';	// backspace
    case 't': reExpression++; return '\t';	// tab
    case 'f': reExpression++; return '\f';	// formfeed
    case 'n': reExpression++; return '\n';	// linefeed
    case 'r': reExpression++; return '\r';	// carriage return

    case '0':				// 0-7 is octal constant
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    {
      unsigned ch = reExpression[0] - '0';
      reExpression++;
      if( *reExpression >= '0' && *reExpression < '8' )
      {
        ch <<= 3;
        ch += reExpression[0] - '0';
        reExpression++;
      }
      if( *reExpression >= '0' && *reExpression < '8')
      {
        ch <<= 3;
        ch += reExpression[0] - '0';
        reExpression++;
      }
      return(ch);
    }
    default:				// otherwise, just that char
      return(reExpression[0]);
  }
}

// ParseClosure() - place closure character and size before the factor
// in the compiled string.

static int ParseClosure(string& factor)
{
  const size_t MaxExprLen = (1U << CHAR_BIT*sizeof(char))-2;

  if( factor.size() > MaxExprLen )
    return RFALSE;			// closure expression too large

  factor.insert(0,factor.size());
  factor.insert(0,reExpression[0]);

  reExpression++;

  return RTRUE;
}

// ParseCCL() - parse and translate a character class. Return pointer to the
// compiled class or NULL on error.

static void ParseCCL( string& ccl )
{
  int first = RTRUE;
  int len;

  ccl = "[ ";

  if( *reExpression == NEGATE )         // if first character is NEGATE (^)
  {
    ccl[0] = NCCL;                  // then we have a negated
    reExpression++;                     // character class
  }

  // parse all characters up to the closing bracket or end of string marker

  while( *reExpression != CCLEND && *reExpression != ENDSTR )
  {
    if( *reExpression == DASH && first == RFALSE )  // DASH, check for range
    {
      reExpression++;
      if( *reExpression == NCCL )
        ccl += DASH;			// not range, literal DASH
      else
      {
        ParseDASH( ccl, *reExpression );
        reExpression++;
      }
    }
    else
    {
      if( *reExpression == ESCAPE )
      {
        reExpression++;
        ccl += ParseEscape();
        reExpression++;
      }
      else
      {
        ccl += reExpression[0];
        reExpression++;
      }
    }
    first = RFALSE;
  }

  len = ccl.size()-2;

  if(len > 255)
    return;			// character class too large
  else
  {                                     // store CCL length at ccl[1]
    ccl[1] = (char)len;
    return;
  }
}

// ParseDASH() - fill in range characters.

static const char *ParseDASH( string& ccl, char ch )
{
  for( char c = ccl[ccl.size() - 1] + 1 ; c <= ch ; c++ )
    ccl += c;

  return ccl.c_str();
}

//----------------------------------------------------------------------------

// Match() - Return the position of the first character of the left-most
// longest substring of s that Matches pattern, or NULL if no match is
// found. Sets RSTART and RLENGTH.

static bool Match( const char *s, const char *pattern, int& rstart, int& rlength )
{
  const char *c = s;
  reExpression  = s;

  while( *c != ENDSTR )
  {
    if( MatchTerm(int(c-s), c, pattern) != RFALSE )
    {
      rstart  = int(c-s);
      rlength = int(reExpression - c);
      return true;
    }
    c++;
  }
  rstart = rlength = 0;
  return false;
}

// MatchTerm() - Match a compiled term. Returns RTRUE, RFALSE, or ALMOST.

static int MatchTerm(int offset, const char* s, const char* pattern)
{
  reExpression = s;

  if(*pattern == ENDSTR)
    return RFALSE;

  do
  {
    switch(*pattern)
    {
      case BOL:				// ^ - match beginning of line
        if(offset != 0)
          return RFALSE;
        pattern++;
        break;

      case LITCHAR:			// c - match literal character
        {
          char c = *s++;
          if(!RegularExpression::CaseSensitive && iswlower(c))
            c = _toupper(c);
          if(c!= *++pattern)
            return RFALSE;
          pattern++;
        }
        break;

      case END_TERM:			// e - skip end-of-term character
        pattern++;
        break;

      case ANY:				// . - match any character
        if(*s++ == ENDSTR)		//     except end of string
          return RFALSE;
        pattern++;
        break;

      case OR:
        return MatchOR(offset, s, pattern);

      case CCL:				// [ - character class
      case NCCL:			// ]
        if(*s == ENDSTR)
          return RFALSE;
        if(!MatchCCL(*s++, pattern++))
          return RFALSE;
        pattern += pattern[0] + 1;
        break;

      case EOL:				// $ - match end of string
        if(*s != ENDSTR)
          return RFALSE;
        pattern++;
        break;

      case ZERO_ONE:			// ?
        return Match_0_1(offset, s, pattern);

      case CLOSURE:			// *
      case POS_CLO:			// +
      {
        string ClosurePattern( pattern+2, *(pattern+1) );
        return MatchClosure(offset, s, pattern, ClosurePattern.c_str());
      }

      default:

      // If we get to this point, then something has gone very wrong.
      // Most likely, someone has tried to match with an invalid
      // compiled pattern.

        return RFALSE;
    }
    reExpression = s;
  }
  while(*pattern != ENDSTR);

  return RTRUE;
}

// MatchOR() - Handles selection processing.

static int MatchOR(int offset, const char* s, const char* pattern)
{
  char tmp_pattern[1024];
  const char *t1, *t2, *junk;

  // The first case is build into tmp_pattern. Second case is already there.
  // Both patterns are searched to determine the longest matched substring.

  tmp_pattern[0] = ENDSTR;
  pattern++;
  junk = SkipTerm(pattern);

  strncat(tmp_pattern, pattern, int(junk-pattern));
  strcat(tmp_pattern, SkipTerm(junk));
  t1 = (MatchTerm(offset, s, tmp_pattern) != RFALSE) ? reExpression : NULL;

  // The second pattern need not be searched if the first pattern results
  // in a match through to the end of the string, since the longest possible
  // match has already been found.

  if(t1 == NULL || *reExpression != ENDSTR)
  {
    t2 = (MatchTerm(offset, s, junk) != RFALSE) ? reExpression : NULL;

    // determine which matched the longest substring

    if(t1 != NULL && (t2 == NULL || t1 > t2))
      reExpression = t1;
  }
  return (t1 == NULL && t2 == NULL) ? RFALSE : RTRUE;
}

// SkipTerm() - Skip over the current term and return a pointer to the
// next term in the pattern.

static const char* SkipTerm(const char* pattern)
{
  register int nterm = 1;

  while(nterm > 0)
  {
    switch(*pattern)
    {
      case OR:
        nterm++;
        break;

      case CCL:
      case NCCL:
      case CLOSURE:
      case ZERO_ONE:
      case POS_CLO:
        pattern++;
        pattern += pattern[0];
        break;

      case END_TERM:
        nterm--;
        break;

      case LITCHAR:
        pattern++;
        break;
    }
    pattern++;
  }
  return pattern;
}

// Match_0_1() - Match the ZERO_ONE operator. First, this routine attempts
// to match the entire pattern with the input string. If that fails, it
// skips over the closure pattern and attempts to match the rest of the
// pattern.

static int Match_0_1(int offset, const char* s, const char* pattern)
{
  const char* old_s = s;

  if(MatchTerm(offset, s, pattern+2) == RTRUE)
    return RTRUE;
  else if(MatchTerm(offset, old_s, pattern+2+*(pattern+1)) == RFALSE)
    return RFALSE;
  else
    return ALMOST;
}

// MatchClosure() - Match CLOSURE and POS_CLO. Match as many of the
// closure patterns as possible, then attempt to match the remaining
// pattern with what's left of the input string. Backtrack until we've
// either matched the remaing pattern or we arrive back at where we
// started.

static int MatchClosure( int offset, const char* s,
                         const char* pattern, const char* ClosurePattern)
{
  const char* old_s = s;

  if(MatchTerm(offset, s, ClosurePattern) == RTRUE)
  {
    old_s = reExpression;
    if(MatchClosure(offset, old_s, pattern, ClosurePattern) != RFALSE)
      return RTRUE;
    else
      return MatchTerm(offset, old_s, pattern+2+*(pattern+1));
  }
  else if(*pattern != CLOSURE)	  // POS_CLO requires at least one match
    return RFALSE;
  else if(MatchTerm(offset, old_s, pattern+2+*(pattern+1)) == RTRUE)
    return ALMOST;
  else
    return RFALSE;
}

// MatchCCL() - Match a character class or negated character class

static int MatchCCL(char c, const char* pattern)
{
  register int x;
  char ccl = *pattern++;

  if(!RegularExpression::CaseSensitive && iswlower(c))
    c = _toupper(c);

  for(x = *pattern; x > 0; x--)
  {
    if(c == pattern[x])
      return(ccl == CCL);
  }
  return(ccl != CCL);
}
