/*****************************************************************************
*   $Id: get.c,v 8.1 1999/03/04 04:16:38 darren Exp $
*
*   Copyright (c) 1996-1999, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains the high level source read functions (preprocessor
*   directives are handled within this level).
// Modified by Mike Melnikov
*****************************************************************************/
#include "stdafx.h"

#include "entry.h"
#include "get.h"
#include "options.h"
#include "read.h"
#include <string>

using namespace std;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*============================================================================
=   Data declarations
============================================================================*/
enum  Comment{ COMMENT_NONE, COMMENT_C, COMMENT_CPLUS };

enum eCppLimits {
  MaxDirectiveName = 10
};

/*============================================================================
=   Data declarations
============================================================================*/

/*  Defines the current state of the pre-processor.
*/
enum eState {
  DRCTV_NONE,		/* no known directive - ignore to end of line */
  DRCTV_DEFINE,	/* "#define" encountered */
  DRCTV_HASH,		/* initial '#' read; determine directive */
  DRCTV_INCLUDE, /* #include encountered */
};

struct  cppState
{
  cppState() : ungetch(0) {}
  int	    ungetch;		/* an ungotten character, if any */
  struct sDirective
  {
    sDirective() : state(DRCTV_NONE),accept(false) {}
    eState 	 state;
    bool     accept;/* is a directive syntatically permitted? */
    string  name;		/* macro name */
  } directive;
};

//  Use brace formatting to detect end of block.
static cppState Cpp;

/*============================================================================
=   Function prototypes
============================================================================*/
static void readDirective (int c, char* const name, unsigned int maxLength);
static void directiveHash (const int c);
static void directiveDefine (const int c);
static void directiveInclude (const int c);
static void handleDirective(const int c);
static Comment isComment();
static int skipOverCComment ();
static int skipOverCplusComment ();
static int skipToEndOfString ();
static int skipToEndOfChar ();

/*============================================================================
=   Function definitions
============================================================================*/

void cppInit()
{
  Cpp.directive.state	    = DRCTV_NONE;
  Cpp.directive.accept    = true;
  ASSERT(Cpp.directive.name.size() == 0);
  Cpp.directive.name.erase();
}

void cppTerminate()
{
  Cpp.directive.name.erase();
}

/*----------------------------------------------------------------------------
*   Scanning functions
*
*   This section handles preprocessor directives.  It strips out all
*   directives and may emit a tag for #define directives.
*--------------------------------------------------------------------------*/

/*  This puts a character back into the input queue for the source File.
*/
void cppUngetc(const int  c )
{
  Cpp.ungetch = c;
}

/*  Reads a directive, whose first character is given by "c", into "name".
*/
static void readDirective(int c,char* const  name,unsigned int  maxLength )
{
  unsigned int i;

  for (i = 0  ;  i < maxLength - 1  ;  ++i)
  {
    if (i > 0)
    {
      c = fileGetc();
      if (c == EOF_FILE  ||  ! isalpha(c))
      {
        fileUngetc();
        break;
      }
    }
    name[i] = c;
  }
  name[i] = '\0';					/* null terminate */
}

/*  Reads an identifier, whose first character is given by "c", into "tag",
*  together with the file location and corresponding line number.
*/
static void readDefineTag(int c,string& name,bool *const  parameterized )
{
  name.erase();
  do
  {
    name +=c;
  } while (c = fileGetc(), (c != EOF_FILE  &&  isident(c)));
  fileUngetc();

  *parameterized = (bool)(c == '(');
}

static void readIncludeTag(int c,string& name )
{
  name.erase();
  do
  {
    name += c;
  } while (c = fileGetc(), (c != EOF_FILE  && 
                            c != DOUBLE_QUOTE &&
                            c != '>'));
  //fileUngetc();
}

static void makeDefineTag(const string& name )
{
  if(Option.include[TAG_DEFINE])
  {
    ClassItem classItem;
    classItem.m_name = name.c_str();
    classItem.m_line = getFileLine();
    writeCtagsEntry(classItem,TAG_DEFINE);
  }
}

extern tagFile TagFile;
static void makeIncludeTag(const string& name )
{
  if(Option.include[TAG_INCLUDE])
  {
    ClassItem classItem;

    // temp realization
    if(Project::commonProject.m_shortNames.size())
    {
        bool found = false;
        string testFile;
        Project* pr = TagFile.m_files[TagFile.currentFile].project; // it can be 'common' project
        found = Project::Normalize(pr->path,name,testFile);
        if(found)
        {
          int idx = TagFile.getFileIndex(testFile);
          if(TagFile.m_files[idx].m_shouldParse == _MaybeParse)
            TagFile.m_files[idx].m_shouldParse = _ShouldParse;
          classItem.m_name = name.c_str();//testFile.c_str();
          found = true;
        }
        if(!found)
        {
          for(int a=0;a<pr->additionalDirs.size();a++)
          {
            found = Project::Normalize(pr->additionalDirs[a],name,testFile);
            if(found)
            {
              int idx = TagFile.getFileIndex(testFile);
              if(TagFile.m_files[idx].m_shouldParse == _MaybeParse)
                TagFile.m_files[idx].m_shouldParse = _ShouldParse;
              classItem.m_name = name.c_str();//testFile.c_str();
              found = true;
              break;
            }
          }
        }

        if(!found)
        {
          Project* pr = &Project::commonProject;
          int count = pr->m_shortNames.size();
          for(int i = 0; i<count;i++)
          {
            if(_stricmp(pr->m_shortNames[i].c_str(),name.c_str())==0)
            {
              classItem.m_name = name.c_str();//pr->m_files[i].c_str();
              int idx = TagFile.getFileIndex(pr->m_files[i]);
              if(TagFile.m_files[idx].m_shouldParse == _MaybeParse)
                TagFile.m_files[idx].m_shouldParse = _ShouldParse;
              found = true;
              break;
            }
          }
        }

        if(!found)
        {
          classItem.m_name = ('\'' + name + '\'').c_str();
        }
    }
    else
      classItem.m_name = ('\'' + name + '\'').c_str();

    classItem.m_line = getFileLine();
    classItem.m_iparents.push_back(TagItem(TAG_FILE,TagFile.currentFile));
    writeCtagsEntry(classItem,TAG_INCLUDE);




  }
}

static void directiveDefine(    const int  c )
{
  bool parameterized;

  if (isident1(c))
  {
    readDefineTag(c, Cpp.directive.name, &parameterized);
    makeDefineTag(Cpp.directive.name);
  }
  Cpp.directive.state = DRCTV_NONE;
}

static void directiveInclude(    const int  c )
{
  //if(isident1(c) || c=='.' || c=='\\')
  {
    readIncludeTag(c, Cpp.directive.name);
    makeIncludeTag(Cpp.directive.name);
  }
  Cpp.directive.state = DRCTV_NONE;
}

static void directiveHash(const int  c )
{
  if(!isdigit(c))
  {
    char directive[MaxDirectiveName];
    readDirective(c, directive, MaxDirectiveName);
    if (strcmp(directive, "define")==0)
      Cpp.directive.state = DRCTV_DEFINE;
    else if (strcmp(directive, "include")==0)
      Cpp.directive.state = DRCTV_INCLUDE;
    else	/* "pragma", "else", "endif" ,"elif", "if" , "undef", etc*/
      Cpp.directive.state = DRCTV_NONE;
  }
}

/*  Handles a pre-processor directive whose first character is given by "c".
*/
static void handleDirective(    const int  c )
{
  switch (Cpp.directive.state)
  {
    case DRCTV_DEFINE:	directiveDefine(c);		break;
    case DRCTV_HASH:	  directiveHash(c);	break;
    case DRCTV_INCLUDE:	directiveInclude(c);		break;
  }
}

/*  Called upon reading of a slash ('/') characters, determines whether a
*  comment is encountered, and its type.
*/
static Comment isComment()
{
  Comment comment;
  const int next = fileGetc();

  if (next == '*')
    comment = COMMENT_C;
  else if (next == '/')
    comment = COMMENT_CPLUS;
  else
  {
    fileUngetc();
    comment = COMMENT_NONE;
  }
  return comment;
}

/*  Skips over a C style comment. According to ANSI specification a comment
*  is treated as white space, so we perform this subsitution.
*/
static int skipOverCComment()
{
  int c = fileGetc();

  while (c != EOF_FILE)
  {
    if (c != '*')
      c = fileGetc();
    else
    {
      const int next = fileGetc();

      if (next != '/')
        c = next;
      else
      {
        c = ' ';			/* replace comment with space */
        break;
      }
    }
  }
  return c;
}

/*  Skips over a C++ style comment.
*/
static int skipOverCplusComment()
{
  int c;

  while ((c = fileGetc()) != EOF_FILE)
  {
    if (c == BACKSLASH)
      fileGetc();			/* throw away next character, too */
    else if (c == NEWLINE)
      break;
  }
  return c;
}

/*  Skips to the end of a string, returning a special character to
*  symbolically represent a generic string.
*/
static int skipToEndOfString()
{
  int c;

  while ((c = fileGetc()) != EOF_FILE)
  {
    if (c == BACKSLASH)
    {
      fileGetc();			/* throw away next character, too */
    }
    else if (c == DOUBLE_QUOTE)
      break;
  }
  return STRING_SYMBOL;		/* symbolic representation of string */
}

/*  Skips to the end of the three (possibly four) 'c' sequence, returning a
*  special character to symbolically represent a generic character.
*/
static int skipToEndOfChar()
{
  int c;

  while ((c = fileGetc()) != EOF_FILE)
  {
    if (c == BACKSLASH)
      fileGetc();			/* throw away next character, too */
    else if (c == SINGLE_QUOTE)
      break;
    else if (c == NEWLINE)
    {
      fileUngetc();
      break;
    }
  }
  return CHAR_SYMBOL;		    /* symbolic representation of character */
}

/*  This function returns the next character, stripping out comments,
*  C pre-processor directives, and the contents of single and double
*  quoted strings. In short, strip anything which places a burden upon
*  the tokenizer.
*/
int cppGetc()
{
  bool directive = false;
  int c;

  if (Cpp.ungetch != '\0')
  {
    c = Cpp.ungetch;
    Cpp.ungetch = '\0';
    return c;
  }
  else do
  {
    c = fileGetc();
    switch (c)
    {
    case EOF_FILE:
      directive = false;
      break;

    case TAB:
    case SPACE:
      break;				/* ignore most white space */

    case NEWLINE:
      directive = false;
      Cpp.directive.accept = true;
      break;

    case DOUBLE_QUOTE:
      if(Cpp.directive.state!=DRCTV_INCLUDE) // was added
      {
        Cpp.directive.accept = false;
		      c = skipToEndOfString();
      }
      break;

    case '<':
      if(Cpp.directive.state==DRCTV_INCLUDE) // skip to get normal name
      break;

    case '#':
      if (Cpp.directive.accept)
      {
        directive = true;
        Cpp.directive.state  = DRCTV_HASH;
        Cpp.directive.accept = false;
      }
      break;

    case SINGLE_QUOTE:
      Cpp.directive.accept = false;
      c = skipToEndOfChar();
      break;

    case '/':
      {
        const Comment comment = isComment();

        if (comment == COMMENT_C)
          c = skipOverCComment();
        else if (comment == COMMENT_CPLUS)
        {
          c = skipOverCplusComment();
          if (c == NEWLINE)
            fileUngetc();
        }
        else
          Cpp.directive.accept = false;
        break;
      }

    case '\\':
      {
        if (fileGetc() == NEWLINE)
          continue;
        else
          fileUngetc();
        break;
      }

    default:
      Cpp.directive.accept = false;
      if (directive)
        handleDirective(c);
      break;
    }
  } while (directive);

  return c;
}
