/*****************************************************************************
*   $Id: keyword.c,v 8.1 1999/03/04 04:16:38 darren Exp $
*
*   Copyright (c) 1998-1999, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   Manages a keyword hash.
// Modified by Mike Melnikov
*****************************************************************************/

#include "stdafx.h"

#include "keyword.h"
#include "options.h"
#include "read.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*============================================================================
=   Macros
============================================================================*/
#define HASH_EXPONENT	7	/* must be less than 17 */

/*============================================================================
=   Data declarations
============================================================================*/
typedef struct sHashEntry {
  const char* str;
  struct sHashEntry *next;
  int value[LANG_COUNT];
} hashEntry;

/*============================================================================
=   Data definitions
============================================================================*/
const unsigned int TableSize = 1 << HASH_EXPONENT;
static hashEntry **HashTable = NULL;

/*============================================================================
=   Function prototypes
============================================================================*/
static unsigned int hashValue (const char* const str);
static hashEntry *newEntry (const char* const str, langType language, int value);

/*============================================================================
=   Function definitions
============================================================================*/
void buildHashTable()
{
  HashTable = new hashEntry*[TableSize];
  for (unsigned int i = 0;i<TableSize;++i)
    HashTable[i] = NULL;
}

static unsigned int hashValue(const char* const  string )
{
  unsigned long value = 0;
  ASSERT(string != NULL);

  /*  We combine the various words of the multiword key using the method
  *  described on page 512 of Vol. 3 of "The Art of Computer Programming".
  */
  for(const unsigned char* p = (const unsigned char* )string  ;  *p != '\0'  ;  ++p)
  {
    value <<= 1;
    if (value & 0x00000100L)
      value = (value & 0x000000ffL) + 1L;
    value ^= *p;
  }
  /*  Algorithm from page 509 of Vol. 3 of "The Art of Computer Programming"
  *  Treats "value" as a 16-bit integer plus 16-bit fraction.
  */
  value *= 40503L;		/* = 2^16 * 0.6180339887 ("golden ratio") */
  value &= 0x0000ffffL;	/* keep fractional part */
  value >>= 16 - HASH_EXPONENT; /* scale up by hash size and move down */

  return value;
}

static hashEntry *newEntry(const char* const  str,langType  language,int  value )
{
  hashEntry *const entry = new hashEntry;
  entry->str = str;
  entry->next = NULL;
  for(int i = 0; i < LANG_COUNT;++i)
    entry->value[i] = 0;
  entry->value[(int)language] = value;
  return entry;
}

/*  Note that it is assumed that a "value" of zero means an undefined keyword
*  and clients of this function should observe this. Also, all keywords added
*  should be added in lower case. If we encounter a case-sensitive language
*  whose keywords are in upper case, we will need to redesign this.
*/
void addKeyword(const char* const  str,langType  language,int value )
{
  const unsigned int hashedValue = hashValue(str);
  ASSERT(hashedValue < TableSize);
  hashEntry *entry = HashTable[hashedValue];

  if (entry == NULL)
  {
    HashTable[hashedValue] = newEntry(str, language, value);
  }
  else
  {
    hashEntry *prev = NULL;

    while (entry != NULL)
    {
      if (strcmp(str, entry->str) == 0)
      {
        entry->value[language] = value;
        break;
      }
      prev = entry;
      entry = entry->next;
    }
    if (entry == NULL)
    {
      hashEntry *newN = newEntry(str, language, value);

      ASSERT(prev != NULL);
      prev->next = newN;
    }
  }
}

int lookupKeyword(const char* const  str)
{
  ASSERT(File.language < LANG_COUNT);

  const unsigned int hashedValue = hashValue(str);
  ASSERT(hashedValue < TableSize);
  hashEntry *entry = HashTable[hashedValue];

  while(entry)
  {
    if (strcmp(str, entry->str) == 0)
      return entry->value[File.language];
    entry = entry->next;
  }
  return 0;
}

void freeKeywordTable()
{
  ASSERT(HashTable);
  for(unsigned int i = 0;i<TableSize;++i)
  {
    hashEntry *entry = HashTable[i];

    while (entry != NULL)
    {
      hashEntry *next = entry->next;
      delete entry;
      entry = next;
    }
  }
  delete HashTable;
}
