/*****************************************************************************
*   $Id: parse.c,v 8.2 1999/05/09 20:07:52 darren Exp $
*
*   Copyright (c) 1996-1999, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains functions for parsing and scanning of a source file.
// Modified by Mike Melnikov
*****************************************************************************/

#include "stdafx.h"

#include "entry.h"
#include "get.h"
#include <string>
#include "options.h"
#include "parse.h"
#include "read.h"
#include "ctags.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*============================================================================
=   Macros
============================================================================*/
#define activeToken(st)		((st)->token[(int)(st)->tokenIndex])
#define isType(token,t)     (bool)((token)->type == (t))
#define isMemberDecl(st)    (bool)((st)->member.isMember)
#define insideEnumBody(st)  (bool)((st)->parent == NULL ? false : \
((st)->parent->declaration == DECL_ENUM))

#define isOneOf(c,s)        (bool)(strchr((s), (c)) != NULL)

/*============================================================================
=   Data declarations
============================================================================*/

#define NumTokens 3

enum  exception_t {
  ExceptionFormattingError,ExceptionBraceFormattingError
};

/*  Used to specify type of keyword.
*/
enum  keywordId{
    KEYWORD_NONE,
    KEYWORD_BOOL, KEYWORD_BYTE,
    KEYWORD_CATCH, KEYWORD_CHAR, KEYWORD_CLASS, KEYWORD_CONST,
    KEYWORD_DOUBLE,
    KEYWORD_ENUM, KEYWORD_EXPLICIT, KEYWORD_EXTERN ,
    KEYWORD_FINAL, KEYWORD_FLOAT, KEYWORD_FRIEND,
    KEYWORD_IMPORT, KEYWORD_INLINE, KEYWORD_INT,KEYWORD_INTERFACE,
    KEYWORD_LONG,
    KEYWORD_MUTABLE,
    KEYWORD_NAMESPACE, KEYWORD_NEW,
    KEYWORD_OPERATOR,
    KEYWORD_PRIVATE, KEYWORD_PROTECTED, KEYWORD_PUBLIC,
    KEYWORD_REGISTER,
    KEYWORD_SHORT, KEYWORD_SIGNED, KEYWORD_STATIC, KEYWORD_STRUCT,
    KEYWORD_TEMPLATE, KEYWORD_THROW, KEYWORD_TRY, KEYWORD_TYPEDEF, KEYWORD_TYPENAME,
    KEYWORD_UNION, KEYWORD_UNSIGNED, KEYWORD_USING,
    KEYWORD_VIRTUAL, KEYWORD_VOID, KEYWORD_VOLATILE,
    KEYWORD_WCHAR_T,
};

/*  Used to determine whether keyword is valid for the current language and
*  what its ID is.
*/
struct  keywordDesc{
  const char* name;
  keywordId id;
  short isValid[LANG_COUNT]; /* indicates languages for which kw is valid */
};

/*  Used for reporting the type of object parsed by nextToken().
*/
enum tokenType{
  TOK_NONE,		/* none */
  TOK_ARGS,		/* a parenthetical pair and its contents */
  TOK_BRACE_CLOSE,
  TOK_BRACE_OPEN,
  TOK_COMMA,		/* the comma character */
  TOK_KEYWORD,
  TOK_NAME,		/* an unknown name */
  TOK_PAREN_NAME,	/* a single name in parentheses */
  TOK_SEMICOLON,	/* the semicolon character */
  TOK_SPEC,		/* a storage class specifier, qualifier, type, etc. */
  TOK_COUNT
};

/*  This describes the scoping of the current statement.
*/
enum  tagScope {
  SCOPE_GLOBAL,		/* no storage class specified */
  SCOPE_STATIC,		/* static storage class */
  SCOPE_EXTERN,       /*  al storage class */
  SCOPE_FRIEND,		/* declares access only */
  SCOPE_TYPEDEF,		/* scoping depends upon context */
  SCOPE_NAMESPACE,  /* scoping namspase */
  SCOPE_COUNT
};

/*  Information about the parent class of a member (if any).
*/
struct  memberInfo{
  bool isMember;   /* is declared entity a member? */
  bool immediate;  /* immediate class name preceding declaration?*/
  accessType	access;		/* access of current statement */
  accessType	accessDefault;	/* access default for current statement */
};

struct  tokenInfo{
  tokenType     type;
  keywordId     keyword;
  string       name;		/* the name of the token */
  unsigned long lineNumber;	/* line number of tag */
};

/*  Describes the statement currently undergoing analysis.
*/
struct  statementInfo {
  tagScope	        scope;
  declType	        declaration;	/* specifier associated with TOK_SPEC */
  bool              gotName;    /* was a name parsed yet? */
  bool              haveQualifyingName;  /* do we have a name we are considering? */
  bool              gotParenName;   /* was a name inside parentheses parsed yet? */
  bool              gotArgs;    /* was a list of parameters parsed yet? */
  bool              isPointer;  /* is 'name' a pointer? */
  unsigned int      tokenIndex;	/* currently active token */
  tokenInfo*	      token[(int)NumTokens];
  tokenInfo*	      context;	/* information on name of current context */
  memberInfo	      member;		/* information regarding parent class/struct */
  statementInfo*    parent;  /* statement we are nested within */
  string            classParents;
};


struct parenInfo{
  bool isPointer;
  bool isParamList;
  bool isKnrParamList;
  bool isNameCandidate;
  bool invalidContents;
  bool nestedArgs;
  unsigned int parameterCount;
};

/*============================================================================
=   Data definitions
============================================================================*/

static const keywordDesc KeywordTable[] = {
  /* keyword		keyword ID		 \   |   /  	*/
  { "bool",     KEYWORD_BOOL,             { 0, 1 } },
  { "byte",		  KEYWORD_BYTE,	            { 0, 1 } },
  { "catch",		KEYWORD_CATCH,		        { 0, 1 } },
  { "char",	  	KEYWORD_CHAR,	            { 0, 1 } },
  { "class",		KEYWORD_CLASS,	        	{ 0, 1 } },
  { "const",		KEYWORD_CONST,		        { 0, 1 } },
  { "double",		KEYWORD_DOUBLE,	        	{ 0, 1 } },
  { "enum",		  KEYWORD_ENUM,		          { 0, 1 } },
  { "explicit",	KEYWORD_EXPLICIT,	        { 0, 1 } },
  { "extern",   KEYWORD_EXTERN,           { 0, 1 } },
  { "float",		KEYWORD_FLOAT,	        	{ 0, 1 } },
  { "friend",		KEYWORD_FRIEND,	        	{ 0, 1 } },
  { "import",		KEYWORD_IMPORT,	        	{ 0, 0 } },
  { "inline",		KEYWORD_INLINE,		        { 0, 1 } },
  { "int",	  	KEYWORD_INT,		          { 0, 1 } },
  { "interface",KEYWORD_INTERFACE,	      { 0, 1 } },
  { "long",	  	KEYWORD_LONG,		          { 0, 1 } },
  { "mutable",	KEYWORD_MUTABLE,        	{ 0, 1 } },
  { "namespace",KEYWORD_NAMESPACE,      	{ 0, 1 } },
  { "new",		  KEYWORD_NEW,		          { 0, 1 } },
  { "operator",	KEYWORD_OPERATOR,       	{ 0, 1 } },
  { "private",	KEYWORD_PRIVATE,        	{ 0, 1 } },
  { "protected",KEYWORD_PROTECTED,	      { 0, 1 } },
  { "public",		KEYWORD_PUBLIC,	        	{ 0, 1 } },
  { "register",	KEYWORD_REGISTER,       	{ 0, 1 } },
  { "short",		KEYWORD_SHORT,	        	{ 0, 1 } },
  { "signed",		KEYWORD_SIGNED,	        	{ 0, 1 } },
  { "static",		KEYWORD_STATIC,	        	{ 0, 1 } },
  { "struct",		KEYWORD_STRUCT,	        	{ 0, 1 } },
  { "template",	KEYWORD_TEMPLATE,	        { 0, 1 } },
  { "throw",		KEYWORD_THROW,		        { 0, 1 } },
  { "try",	  	KEYWORD_TRY,		          { 0, 1 } },
  { "typedef",	KEYWORD_TYPEDEF,	        { 0, 1 } },
  { "typename",	KEYWORD_TYPENAME,	        { 0, 1 } },
  { "union",		KEYWORD_UNION,		        { 0, 1 } },
  { "unsigned",	KEYWORD_UNSIGNED,	        { 0, 1 } },
  { "using",		KEYWORD_USING,		        { 0, 1 } },
  { "virtual",	KEYWORD_VIRTUAL,	        { 0, 1 } },
  { "void",	  	KEYWORD_VOID,		          { 0, 1 } },
  { "volatile",	KEYWORD_VOLATILE,	        { 0, 1 } },
  { "wchar_t",	KEYWORD_WCHAR_T,	        { 0, 1 } },
};

/*============================================================================
=   Function prototypes
============================================================================*/
static void initToken (tokenInfo *const token);
static void advanceToken (statementInfo *const st);
static void retardToken (statementInfo *const st);
static tokenInfo *prevToken (const statementInfo *const st, unsigned int n);
static void setToken (statementInfo *const st, const tokenType type);
static tokenInfo *newToken();
static void reinitStatement (statementInfo *const st, const bool comma);
static void initMemberInfo (statementInfo *const st);
static void initStatement (statementInfo *const st, statementInfo *const parent);
static tagType declToTagType (const declType declaration);
static void addOtherFields(const tagType type, const statementInfo *const st, string& scope, ClassItem& e);
static void makeTag(const tokenInfo *const token, const statementInfo *const st, const tagType type, bool negative = false);
static bool isValidTypeSpecifier (const declType declaration);
static void qualifyEnumeratorTag (const statementInfo *const st, const tokenInfo *const nameToken);
static void qualifyFunctionDeclTag (const statementInfo *const st, const tokenInfo *const nameToken);
static void qualifyCompoundTag (const statementInfo *const st, const tokenInfo *const nameToken);
static void qualifyBlockTag (const statementInfo *const st, const tokenInfo *const nameToken);
static void qualifyVariableTag (const statementInfo *const st, const tokenInfo *const nameToken);
static int skipToCharacter (const int target);
static void skipToMatch (const char* const _pair);
static void skipParens();
static void analyzeIdentifier (tokenInfo *const token);
static void readIdentifier (tokenInfo *const token, const int firstChar);
static void readOperator (statementInfo *const st);
static void copyToken (tokenInfo *const dest, const tokenInfo *const src);
static void setAccess (statementInfo *const st, const accessType access);
static void processName (statementInfo *const st);
static void processToken (tokenInfo *const token, statementInfo *const st);
static void restartStatement (statementInfo *const st);
static void skipMemIntializerList (tokenInfo *const token);
static void skipMacro (statementInfo *const st);
static bool skipPostArgumentStuff (statementInfo *const st, parenInfo *const info);
static void analyzePostParens (statementInfo *const st, parenInfo *const info);
static void parseParens (statementInfo *const st, parenInfo *const info);
static void initParenInfo (parenInfo *const info);
static void analyzeParens (statementInfo *const st);
static void addContext (statementInfo *const st);
static void processColon (statementInfo *const st);
static int skipInitializer (const bool enumInitializer);
static void processInitializer (statementInfo *const st);
static void parseIdentifier (statementInfo *const st, const int c);
static bool nextToken (statementInfo *const st);
static statementInfo *newStatement (statementInfo *const parent);
static void deleteStatement();
static void deleteAllStatements();
static bool isContextCandidate (const statementInfo *const st);
static bool isStatementEnd (const statementInfo *const st);
static void checkStatementEnd (statementInfo *const st);
static void nest (statementInfo *const st, const unsigned int nestLevel);
static void tagCheck (statementInfo *const st);
static void createTags (const unsigned int nestLevel, statementInfo *const parent);
static void init();

/*============================================================================
=   Function definitions
============================================================================*/

/*----------------------------------------------------------------------------
*-	Token management
----------------------------------------------------------------------------*/

static void initToken(tokenInfo *const  token )
{
  token->type		= TOK_NONE;
  token->keyword	= KEYWORD_NONE;
  token->lineNumber	= getFileLine();
  token->name.erase();
}

static void advanceToken(statementInfo *const  st )
{
  if (st->tokenIndex >= (unsigned int)NumTokens - 1)
    st->tokenIndex = 0;
  else
    ++st->tokenIndex;

  initToken(st->token[st->tokenIndex]);
}

static void retardToken(statementInfo *const  st )
{
  if (st->tokenIndex == 0)
    st->tokenIndex = (unsigned int)NumTokens - 1;
  else
    --st->tokenIndex;
  setToken(st, TOK_NONE);
}

static tokenInfo *prevToken(const statementInfo *const  st,unsigned int  n )
{
  unsigned int tokenIndex;
  unsigned int num = (unsigned int)NumTokens;

  ASSERT(n < num);
  tokenIndex = (st->tokenIndex + num - n) % num;

  return st->token[tokenIndex];
}

static void setToken(statementInfo *const  st, const tokenType type )
{
  tokenInfo *token;

  token = activeToken(st);
  initToken(token);
  token->type = type;
}

static tokenInfo *newToken()
{
  tokenInfo *const token = new tokenInfo;
  initToken(token);
  return token;
}

/*----------------------------------------------------------------------------
*-	Statement management
----------------------------------------------------------------------------*/

static void reinitStatement(statementInfo *const  st,const bool  partial )
{
  unsigned int i;

  if (! partial)
  {
    st->scope	= (st->parent && st->parent->declaration==DECL_NAMESPACE) ?
                                                    SCOPE_NAMESPACE : SCOPE_GLOBAL;
    if (st->member.isMember)
      st->declaration = DECL_BASE;
    else
      st->declaration = DECL_NONE;
  }
  st->gotParenName    = false;
  st->isPointer   = false;
  st->gotArgs     = false;
  st->gotName     = false;
  st->haveQualifyingName = false;
  st->tokenIndex	= 0;

  for (i = 0  ;  i < (unsigned int)NumTokens  ;  ++i)
    initToken(st->token[i]);

  initToken(st->context);

  /*  Init member info.
  */
  if (! partial)
    st->member.access = st->member.accessDefault;
  if (st->member.immediate)
  {
    st->member.isMember  = false;
    st->member.immediate = false;
  }

  st->classParents.erase();
}

static void initMemberInfo(statementInfo *const  st )
{
  /*const */statementInfo* parent = st->parent;
  accessType accessDefault = ACCESS_UNDEFINED;
  bool isMember = false;

  if (parent != NULL) switch (parent->declaration)
  {
  case DECL_ENUM:
  case DECL_NAMESPACE:
  case DECL_UNION:
    isMember = true;
    accessDefault = ACCESS_UNDEFINED;
    break;

  case DECL_CLASS:
    isMember = true;
    accessDefault = ACCESS_PRIVATE;
    break;

  case DECL_INTERFACE:
  case DECL_STRUCT:
    isMember = true;
    accessDefault = ACCESS_PUBLIC;
    break;

  default: break;
  }
  st->member.isMember	     = isMember;
  st->member.accessDefault = accessDefault;
  st->member.access	     = accessDefault;
  st->member.immediate     = false;
}

static void initStatement(statementInfo *const  st,statementInfo *const  parent )
{
  st->parent = parent;
  initMemberInfo(st);
  reinitStatement(st, false);
}

static tagType declToTagType(const declType  declaration )
{
  tagType type = TAG_UNDEFINED;
  switch (declaration)
  {
    case DECL_CLASS:	  type = TAG_CLASS;	break;
    case DECL_ENUM:		  type = TAG_ENUM;	break;
    case DECL_FUNCTION:	type = TAG_FUNCTION;	break;
    case DECL_INTERFACE:type = TAG_INTERFACE;	break;
    case DECL_STRUCT:	  type = TAG_STRUCT;	break;
    case DECL_UNION:	  type = TAG_UNION;	break;
    default: ASSERT("unexpected declaration" == NULL); break;
  }
  return type;
}

static void addOtherFields(const tagType type,const statementInfo *const st,string& scope,ClassItem& classItem)
{
/*  For selected tag types, append an extension flag designating the
*  parent object in which the tag is defined.
  */
  if (isMemberDecl(st)) switch (type)
  {
  default: break;

  case TAG_CLASS:
  case TAG_ENUM:
  case TAG_ENUMERATOR:
  case TAG_FUNCTION:
  case TAG_MEMBER:
  case TAG_STRUCT:
  case TAG_TYPEDEF:
  case TAG_UNION:
  case TAG_VARIABLE:
    //if (! (type == TAG_ENUMERATOR  &&  vStringLength(scope) == 0))
    if (scope.size())
    {
      classItem.parents = scope.c_str();
    }
    /*
    if (Option.access)
    {
      if (st->scope == SCOPE_FRIEND)
      {
        classItem.access = ACCESS_FRIEND;
      }
      classItem.access = st->member.access;
    }
    */
    break;
  }
}

static void findScopeHierarchy(string& str,const statementInfo *const  st )
{
  const statementInfo *s = st;
  string temp;
  bool first = true;
  unsigned depth = 0;

  str.erase();
  while (s->member.isMember  &&  s->parent != NULL)
  {
    if(s->parent->declaration!=DECL_NAMESPACE)
    {
      string& name = s->parent->context->name;
      if(name.size())
        str = name;
      if(first)
        first = false;
      else
      {
        str += '\\';
      }
      str += temp;
      ++depth;
    }
    s = s->parent;
  }

  /*  We don't specify "<unnamed>" by itself.
  */
  /*if(depth <= 1  &&  vStringLength(st->parent->context->name) == 0)
  {
    ASSERT(vStringLength(str)==0);
    vStringClear(str);
  } */
  ASSERT(!(depth<= 1&&st->parent->context->name.size()==0) || (str.size()==0));
}

static void makeTag(const tokenInfo *const token,const statementInfo *const st,const tagType type, bool negative)
{
/*  Nothing is really of file scope when it appears in a header file.
  */
  ASSERT(type>=0 && type<TAG_COUNT);
  if(isType(token, TOK_NAME)  &&  token->name.size() > 0  &&
    Option.include[type])
  {
    string scope;

    if (st->member.isMember)
    {
      if (st->member.immediate)
        scope=st->context->name;
      else
        findScopeHierarchy(scope, st);
    }

    ClassItem classItem;
    classItem.m_name = token->name.c_str();
    classItem.m_line = negative ? -token->lineNumber : token->lineNumber;
    classItem.classParents = st->classParents.c_str();

    addOtherFields(type, st, scope,classItem);

    writeCtagsEntry(classItem,type);
  }
}

static bool isValidTypeSpecifier(const declType  declaration )
{
  switch (declaration)
  {
    case DECL_BASE:
    case DECL_CLASS:
    case DECL_ENUM:
    case DECL_STRUCT:
    case DECL_UNION:
      return true;
  }
  return false;
}

static void qualifyEnumeratorTag(const statementInfo *const  st,const tokenInfo *const  nameToken )
{
  if (isType(nameToken, TOK_NAME))
    makeTag(nameToken, st, TAG_ENUMERATOR);
}

static void qualifyFunctionDeclTag(const statementInfo *const  st,const tokenInfo *const  nameToken )
{
  if (st->scope == SCOPE_TYPEDEF)
    makeTag(nameToken, st, TAG_TYPEDEF);
  else if (isValidTypeSpecifier(st->declaration))
    makeTag(nameToken, st, TAG_FUNCTION,true); // TAG_PROTOTYPE
}

static void qualifyCompoundTag(const statementInfo *const  st,const tokenInfo *const  nameToken )
{
  const tagType type = declToTagType(st->declaration);
  ASSERT(type != TAG_UNDEFINED);
  makeTag(nameToken, st, type);
}

static void qualifyBlockTag(const statementInfo *const  st,const tokenInfo *const  nameToken )
{
  switch (st->declaration)
  {
 	case DECL_NAMESPACE:
    break;
  case DECL_CLASS:
  case DECL_ENUM:
  case DECL_INTERFACE:
  case DECL_STRUCT:
  case DECL_UNION:
  case DECL_FUNCTION:
    qualifyCompoundTag(st, nameToken);
    break;

  default: break;
  }
}

static void qualifyVariableTag(const statementInfo *const  st,const tokenInfo *const  nameToken )
{
/*	We have to watch that we do not interpret a declaration of the
*	form "struct tag;" as a variable definition. In such a case, the
*	token preceding the name will be a keyword.
  */
  if (st->declaration == DECL_IGNORE)
    ;
  else if (st->scope == SCOPE_TYPEDEF)
    makeTag(nameToken, st, TAG_TYPEDEF);
  else if (isValidTypeSpecifier(st->declaration))
  {
    if (isMemberDecl(st))
    {
      if (st->scope == SCOPE_GLOBAL  ||  st->scope == SCOPE_STATIC)
        makeTag(nameToken, st, TAG_MEMBER);
    }
    else
    {
      if (st->scope == SCOPE_NAMESPACE || st->scope == SCOPE_EXTERN ||  ! st->haveQualifyingName)
        makeTag(nameToken, st, TAG_EXTERN_VAR);
      else
        makeTag(nameToken, st, TAG_VARIABLE);
    }
  }
}

/*----------------------------------------------------------------------------
*-	Parsing functions
----------------------------------------------------------------------------*/

/*  Skip to the next non-white character.
*/
static int skipToCharacter(const int  target )
{
  int c;
  do
  c = cppGetc();
  while (c != target  &&  c != EOF_FILE);
  return c;
}

/*  Skip to the next non-white character.
*/
int skipToNonWhite()
{
  int c;
  do
    c = cppGetc();
  while ((c==SPACE || c==TAB || c==NEWLINE || c==CRETURN) /* && c!=EOF_FILE*/);
  return c;
}

/*  Skip to the matching character indicated by the pair string. If skipping
*  to a matching brace and any brace is found within a different level of a
*  #if conditional statement while brace formatting is in effect, we skip to
*  the brace matched by its formatting. It is assumed that we have already
*  read the character which starts the group (i.e. the first character of
*  "pair").
*/
static void skipToMatch(const char* const  _pair )
{
  const bool braceMatching = (bool)(strcmp("{}", _pair) == 0);
  const int begin = _pair[0], end = _pair[1];
  int matchLevel = 1;
  int c = '\0';

  while (matchLevel > 0  &&  (c = cppGetc()) != EOF_FILE)
  {
    if (c == begin)
    {
      ++matchLevel;
    }
    else if (c == end)
    {
      --matchLevel;
    }
  }
  if (c == EOF_FILE)
  {
    AfxTrace("%s(%ld): failed to find match for '%c'",getFileName(),getFileLine(),begin);
    if (braceMatching)
      throw(ExceptionBraceFormattingError);
    else
      throw(ExceptionFormattingError);
  }
}

static void skipParens()
{
  const int c = skipToNonWhite();

  if (c == '(')
    skipToMatch("()");
  else
    cppUngetc(c);
}

/*  Analyzes the identifier contained in a statement described by the
*  statement structure and adjusts the structure according the significance
*  of the identifier.
*/
static void analyzeIdentifier(tokenInfo *const  token )
{
  ///const char* replacement = NULL;
  bool parensToo = false;

  if(!isIgnoreToken(token->name.c_str(),parensToo/*, &replacement*/))
  {
    /*  Analyzes the identifier contained in a statement described by the
    *  statement structure and adjusts the structure according the significance
    *  of the identifier.
    */
    ///if (replacement != NULL)
      ///token->keyword = (keywordId)lookupKeyword(replacement);
    ///else
      token->keyword = (keywordId)lookupKeyword(token->name.c_str());

    if (token->keyword == KEYWORD_NONE)
      token->type = TOK_NAME;
    else
      token->type = TOK_KEYWORD;
  }
  else
  {
    initToken(token);
    if (parensToo)
    {
      skipParens();
    }
  }
}

static void readIdentifier(tokenInfo *const  token,const int  firstChar )
{
  string& name = token->name;
  char c = firstChar;

  initToken(token);

  do
  {
    name += c;
    c = cppGetc();
  } while (isident(c));
  cppUngetc(c);		/* unget non-identifier character */

  analyzeIdentifier(token);
}

static void readOperator(statementInfo *const  st )
{
  const char* const acceptable = "+-*/%^&|~!=<>,[]";
  tokenInfo *const token = activeToken(st);
  string& name = token->name;
  int c = skipToNonWhite();

  /*  When we arrive here, we have the keyword "operator" in 'name'.
  */
  if (c == '(')
  {
  /*  Verify whether this is a valid "()" operator.
    */
    if (cppGetc() == ')')
    {
      c = skipToNonWhite();
      if (c == '(')
        name += "()";
    }
    else
    {
      skipToMatch("()");
      c = cppGetc();
    }
  }
  else if (isident1(c))
  {
    // commented by zmike
//    if (c == 'd'  ||  c == 'n')	/* only "delete" or "new" permitted */
    {
      name += ' ';
      do
      {
        name += c;
        c = cppGetc();
      //} while (isident(c));
      } while (isidentOperator(c));
      //Strip trailing spaces from str (name);
    }
  }
  else if (isOneOf(c, acceptable))
  {
    name += ' ';
    do
    {
      name += c;
      c = cppGetc();
    } while (isOneOf(c, acceptable));
  }
  cppUngetc(c);

  token->type	= TOK_NAME;
  token->keyword = KEYWORD_NONE;
  processName(st);
}

static void copyToken(tokenInfo *const  dest,const tokenInfo *const  src )
{
  dest->type         = src->type;
  dest->keyword      = src->keyword;
  dest->lineNumber   = src->lineNumber;
  dest->name         = src->name;
}

static void setAccess(statementInfo *const  st,const accessType  access )
{
  if (isMemberDecl(st))
  {
    st->member.accessDefault = access;
    st->member.access = access;
  }
}

static void processName(statementInfo *const  st )
{
  ASSERT(isType(activeToken(st), TOK_NAME));
  if (st->gotName  &&  st->declaration == DECL_NONE)
    st->declaration = DECL_BASE;
  st->gotName = true;
  st->haveQualifyingName = true;
}

static void processToken(tokenInfo *const  token,statementInfo *const  st )
{
  switch (token->keyword)		/* is it a reserved word? */
  {
  default: break;

  case KEYWORD_NONE:	processName(st);			break;
  case KEYWORD_BOOL:  st->declaration = DECL_BASE;		break;
  case KEYWORD_BYTE:  st->declaration = DECL_BASE;		break;
  case KEYWORD_CATCH:	skipParens();				break;
  case KEYWORD_CHAR:	st->declaration = DECL_BASE;		break;
  case KEYWORD_CLASS:	st->declaration = DECL_CLASS;		break;
  case KEYWORD_CONST:	st->declaration = DECL_BASE;		break;
  case KEYWORD_DOUBLE:	st->declaration = DECL_BASE;		break;
  case KEYWORD_ENUM:	st->declaration = DECL_ENUM;		break;
  case KEYWORD_FLOAT:	st->declaration = DECL_BASE;		break;
  case KEYWORD_FRIEND:	st->scope	= SCOPE_FRIEND;		break;
  case KEYWORD_IMPORT:	st->declaration = DECL_IGNORE;		break;
  case KEYWORD_INT:	st->declaration = DECL_BASE;		break;
  case KEYWORD_INTERFACE: st->declaration = DECL_INTERFACE;	break;
  case KEYWORD_LONG:	st->declaration = DECL_BASE;		break;
  case KEYWORD_NAMESPACE: st->declaration = DECL_NAMESPACE;
    st->scope = SCOPE_NAMESPACE;
    break;
  case KEYWORD_OPERATOR:	readOperator(st);			break;
  case KEYWORD_PRIVATE:	setAccess(st, ACCESS_PRIVATE);		break;
  case KEYWORD_PROTECTED:	setAccess(st, ACCESS_PROTECTED);	break;
  case KEYWORD_PUBLIC:	setAccess(st, ACCESS_PUBLIC);		break;
  case KEYWORD_SHORT:	st->declaration = DECL_BASE;		break;
  case KEYWORD_SIGNED:	st->declaration = DECL_BASE;		break;
  case KEYWORD_STRUCT:	st->declaration = DECL_STRUCT;		break;
  case KEYWORD_TYPEDEF:	st->scope	= SCOPE_TYPEDEF;	break;
  case KEYWORD_UNION:	st->declaration = DECL_UNION;		break;
  case KEYWORD_UNSIGNED:	st->declaration = DECL_BASE;		break;
  case KEYWORD_USING:	st->declaration = DECL_IGNORE;		break;
  case KEYWORD_VOID:	st->declaration = DECL_BASE;		break;
  case KEYWORD_VOLATILE:	st->declaration = DECL_BASE;		break;

  case KEYWORD_EXTERN:
    st->scope = SCOPE_EXTERN ;
    st->declaration = DECL_BASE;
    break;

  case KEYWORD_STATIC:
  		st->scope = SCOPE_STATIC;
      st->declaration = DECL_BASE;
      break;
  }
}

/*----------------------------------------------------------------------------
*-	Parenthesis handling functions
----------------------------------------------------------------------------*/

static void restartStatement(statementInfo *const  st )
{
  tokenInfo *const save = newToken();
  tokenInfo *token = activeToken(st);

  copyToken(save, token);
  reinitStatement(st, false);
  token = activeToken(st);
  copyToken(token, save);
  delete save;
  processToken(token, st);
}

/*  Skips over a the mem-initializer-list of a ctor-initializer, defined as:
*
*  mem-initializer-list:
*    mem-initializer, mem-initializer-list
*
*  mem-initializer:
*    [::] [nested-name-spec] class-name (...)
*    identifier
*/
static void skipMemIntializerList(tokenInfo *const  token )
{
  int c;

  do
  {
    c = skipToNonWhite();
    while (isident1(c)  ||  c == ':')
    {
      if (c != ':')
        readIdentifier(token, c);
      c = skipToNonWhite();
    }
    if (c == '<')
    {
      skipToMatch("<>");
      c = skipToNonWhite();
    }
    if (c == '(')
    {
      skipToMatch("()");
      c = skipToNonWhite();
    }
  } while (c == ',');
  cppUngetc(c);
}

static void skipMacro(statementInfo *const  st )
{
  tokenInfo *const prev2 = prevToken(st, 2);

  if (isType(prev2, TOK_NAME))
    retardToken(st);
  skipToMatch("()");
}

/*  Skips over characters following the parameter list. This will be either
*  non-ANSI style function declarations or C++ stuff. Our choices:
*
*  C (K&R):
*    int func();
*    int func (one, two) int one; float two; {...}
*  C (ANSI):
*    int func (int one, float two);
*    int func (int one, float two) {...}
*  C++:
*    int foo (...) [const|volatile] [throw (...)];
*    int foo (...) [const|volatile] [throw (...)] [ctor-initializer] {...}
*    int foo (...) [const|volatile] [throw (...)] try [ctor-initializer] {...}
*        catch (...) {...}
*/
static bool skipPostArgumentStuff(statementInfo *const  st,parenInfo *const  info )
{
  tokenInfo *const token = activeToken(st);
  unsigned int parameters = info->parameterCount;
  unsigned int elementCount = 0;
  bool restart = false;
  bool end = false;
  int c = skipToNonWhite();

  do
  {
    switch (c)
    {
    case ')':				break;
    case ':': skipMemIntializerList(token);	break;	/* ctor-initializer */
    case '[': skipToMatch("[]");		break;
    case '{': cppUngetc(c); end = true; break;
    case '}': cppUngetc(c); end = true; break;

    case '(':
      if (elementCount > 0)
        ++elementCount;
      skipToMatch("()");
      break;

    case ';':
      if (parameters == 0  ||  elementCount < 2)
      {
        cppUngetc(c);
        end = true;
      }
      else if (--parameters == 0)
        end = true;
      break;

    default:
      if (isident1(c))
      {
        readIdentifier(token, c);
        switch (token->keyword)
        {
        case KEYWORD_THROW:	skipParens();	break;
        case KEYWORD_CONST:			break;
        case KEYWORD_TRY:			break;
        case KEYWORD_VOLATILE:			break;

        case KEYWORD_CATCH:	case KEYWORD_CLASS:
        case KEYWORD_EXPLICIT:  case KEYWORD_EXTERN :
        case KEYWORD_FRIEND:	case KEYWORD_INLINE:
        case KEYWORD_MUTABLE:	case KEYWORD_NAMESPACE:
        case KEYWORD_NEW:	case KEYWORD_OPERATOR:
        case KEYWORD_PRIVATE:
        case KEYWORD_PROTECTED:	case KEYWORD_PUBLIC:
        case KEYWORD_STATIC:	case KEYWORD_TEMPLATE:
        case KEYWORD_TYPEDEF:	case KEYWORD_TYPENAME:
        case KEYWORD_USING:	case KEYWORD_VIRTUAL:
        /*  Never allowed within parameter declarations.
          */
          restart = true;
          end = true;
          break;

        default:
          if (isType(token, TOK_NONE))
            ;
          else if (info->isKnrParamList  &&  info->parameterCount > 0)
            ++elementCount;
          else
          {
          /*  If we encounter any other identifier immediately
          *  following an empty parameter list, this is almost
          *  certainly one of those Microsoft macro "thingies"
          *  that the automatic source code generation sticks
          *  in. Terminate the current statement.
            */
            restart = true;
            end = true;
          }
          break;
        }
      }
    }
    if (! end)
    {
      c = skipToNonWhite();
      if (c == EOF_FILE)
        end = true;
    }
  } while (! end);

  if (restart)
    restartStatement(st);
  else
    setToken(st, TOK_NONE);

  return (bool)(c != EOF_FILE);
}

static void analyzePostParens(statementInfo *const  st,parenInfo *const  info )
{
  const unsigned long lineNumber = getFileLine();
  int c = skipToNonWhite();

  cppUngetc(c);
  if (isOneOf(c, "{;,="))
    ;
  else
  {
    if (! skipPostArgumentStuff(st, info))
    {
      AfxTrace("%s(%ld): end of file afer function declaration?",getFileName(), lineNumber);
      throw(ExceptionFormattingError);
    }
  }
}

static void parseParens(statementInfo *const  st,parenInfo *const  info )
{
  tokenInfo *const token = activeToken(st);
  unsigned int identifierCount = 0;
  unsigned int depth = 1;
  bool firstChar = true;

  info->parameterCount = 1;
  do
  {
    int c = skipToNonWhite();

    switch (c)
    {
    case '&':
    case '*':
      info->isPointer = true;
      info->isKnrParamList = false;
      initToken(token);
      break;

    case ':':
      info->isKnrParamList = false;
      break;

    case '.':
      info->isNameCandidate = false;
      info->isKnrParamList = false;
      break;

    case ',':
      info->isNameCandidate = false;
      if (info->isKnrParamList)
      {
        ++info->parameterCount;
        identifierCount = 0;
      }
      break;

    case '=':
      info->isKnrParamList = false;
      info->isNameCandidate = false;
      if (firstChar)
      {
        info->isParamList = false;
        skipMacro(st);
        depth = 0;
      }
      break;

    case '[':
      info->isKnrParamList = false;
      skipToMatch("[]");
      break;

    case '<':
      info->isKnrParamList = false;
      skipToMatch("<>");
      break;

    case ')':
      if (firstChar)
        info->parameterCount = 0;
      --depth;
      break;

    case '(':
      info->isKnrParamList = false;
      if (firstChar)
      {
        info->isNameCandidate = false;
        cppUngetc(c);
        skipMacro(st);
        depth = 0;
      }
      else if (isType(token, TOK_PAREN_NAME))
      {
        cppUngetc(c);
        info->nestedArgs = true;
      }
      else
        ++depth;
      break;

    default:
      if (isident1(c))
      {
        if (++identifierCount > 1)
          info->isKnrParamList = false;
        readIdentifier(token, c);
        if (isType(token, TOK_NAME)  &&  info->isNameCandidate)
          token->type = TOK_PAREN_NAME;
        else if (isType(token, TOK_KEYWORD))
        {
          info->isKnrParamList = false;
          info->isNameCandidate = false;
        }
      }
      else
      {
        info->isParamList     = false;
        info->isKnrParamList  = false;
        info->isNameCandidate = false;
        info->invalidContents = true;
      }
      break;
    }
    firstChar = false;
  } while (! info->nestedArgs  &&  depth > 0  &&
    (info->isKnrParamList  ||  info->isNameCandidate));

  if (! info->nestedArgs) while (depth > 0)
  {
    skipToMatch("()");
    --depth;
  }

  if (! info->isNameCandidate)
    initToken(token);
}

static void initParenInfo(parenInfo *const  info )
{
  info->isPointer     = false;
  info->isParamList       = true;
  info->isKnrParamList    = true;
  info->isNameCandidate   = true;
  info->invalidContents   = false;
  info->nestedArgs        = false;
  info->parameterCount	= 0;
}

static void analyzeParens(statementInfo *const  st )
{
  tokenInfo *const prev = prevToken(st, 1);

  if (! isType(prev, TOK_NONE))    /* in case of ignored enclosing macros */
  {
    tokenInfo *const token = activeToken(st);
    parenInfo info;
    int c;

    initParenInfo(&info);
    parseParens(st, &info);
    c = skipToNonWhite();
    cppUngetc(c);
    if (info.invalidContents)
      reinitStatement(st, false);
    else if (info.isNameCandidate  &&  isType(token, TOK_PAREN_NAME)  &&
      ! st->gotParenName  &&
      (! st->haveQualifyingName  ||  isOneOf(c, "(=")  ||
      (st->declaration == DECL_NONE  &&  isOneOf(c, ",;"))))
    {
      token->type = TOK_NAME;
      processName(st);
      st->gotParenName = true;
      if (! (c == '('  &&  info.nestedArgs))
        st->isPointer = info.isPointer;
    }
    else if (! st->gotArgs  &&  info.isParamList)
    {
      st->gotArgs = true;
      setToken(st, TOK_ARGS);
      advanceToken(st);
      analyzePostParens(st, &info);
    }
    else
      setToken(st, TOK_NONE);
  }
}

/*----------------------------------------------------------------------------
*-	Token parsing functions
----------------------------------------------------------------------------*/

static void addContext(statementInfo *const  st )
{
  tokenInfo *const nameToken = prevToken(st, 1);

  if (st->context->name.size())
    st->context->name += '\\';
  st->context->name += nameToken->name;
  st->haveQualifyingName = false;
  st->member.isMember    = true;
  st->member.immediate   = true;
  if (st->declaration == DECL_NONE)
    st->declaration = DECL_BASE;
}

static void processColon(statementInfo *const  st )
{
  static char nowname[500];

  int c = skipToNonWhite();
  const bool doubleColon = (bool)(c == ':');

  if (! doubleColon)
  {
    cppUngetc(c);
    string& name = st->classParents;
    if(st->declaration == DECL_CLASS || st->declaration == DECL_STRUCT || st->declaration == DECL_INTERFACE)
    {
      do
      {
        c = skipToNonWhite();
        switch (c)
        {
          case EOF_FILE:   break;
          case ',':   name += '\\';break;
          case '<':   skipToMatch("<>");break;
          case '{':   break;

          default:
            if (isident(c))
            {
              char* nname = nowname;
              do
              {
                *nname++ = c;
                c = cppGetc();
              } while (isident(c));
              *nname++ = '\0';
              int qqq = lookupKeyword(nowname);
              switch(qqq)
              {
                case KEYWORD_VIRTUAL :
                case KEYWORD_PRIVATE:	case KEYWORD_PROTECTED:	case KEYWORD_PUBLIC:
                  break;
                default:
                  name += nowname;
              }
              cppUngetc(c);		/* unget non-identifier character */
            }
        }
      } while (c!='{' && c!=EOF_FILE);
      //const int c = skipToCharacter('{');
      cppUngetc(c);
    }
  }

  /*  If we are not inside of a class (or other composite object, then
  *  this is either a method or static member definition. Otherwise it
  *  is a qualified type declaration.
  */
  else if (st->parent == NULL || st->parent->declaration==DECL_NAMESPACE)
    addContext(st);
}

/*  Skips over any initializing value which may follow an '=' character in a
*  variable definition.
*/
static int skipInitializer(const bool  enumInitializer )
{
  bool done = false;
  int c;

  while (! done)
  {
    c = cppGetc();

    if (c == EOF_FILE)
      throw(ExceptionFormattingError);
    else switch (c)
    {
      case ',':
      case ';':   done = true; break;

      case '[':   skipToMatch("[]"); break;
      case '(':   skipToMatch("()"); break;
      case '{':   skipToMatch("{}"); break;
      case '}':
        if (enumInitializer)
          done = true;
        else
        {
          AfxTrace("%s(%ld): unexpected closing brace",getFileName(), getFileLine());
          throw(ExceptionBraceFormattingError);
        }
        break;

      default:    break;
    }
  }
  return c;
}

static void processInitializer(statementInfo *const  st )
{
  const bool inEnumBody = insideEnumBody(st);
  const int c = skipInitializer(inEnumBody);

  if (c == ';')
    setToken(st, TOK_SEMICOLON);
  else if (c == ',')
    setToken(st, TOK_COMMA);
  else if ('}'  &&  inEnumBody)
  {
    cppUngetc(c);
    setToken(st, TOK_COMMA);
  }
  if (st->scope == SCOPE_EXTERN )
    st->scope = SCOPE_GLOBAL;
}

static void parseIdentifier(statementInfo *const  st,const int  c )
{
  tokenInfo *const token = activeToken(st);

  readIdentifier(token, c);
  if (! isType(token, TOK_NONE))
    processToken(token, st);
}

/*  Reads characters from the pre-processor and assembles tokens, setting
*  the current statement state.
*/
static bool nextToken(statementInfo *const  st )
{
  tokenInfo *token = activeToken(st);

  do
  {
    int c = skipToNonWhite();

    switch (c)
    {
    case EOF_FILE:   return false;		break;
    case '(':   analyzeParens(st);  token = activeToken(st);	break;
    case '*':   st->haveQualifyingName = false;         break;
    case ',':   setToken(st, TOK_COMMA);			break;
    case ':':   processColon(st);				break;
    case ';':   setToken(st, TOK_SEMICOLON);			break;
    case '<':   skipToMatch("<>");				break;
    case '=':   processInitializer(st);				break;
    case '[':   skipToMatch("[]");				break;
    case '{':   setToken(st, TOK_BRACE_OPEN);			break;
    case '}':   setToken(st, TOK_BRACE_CLOSE);			break;

    default:
      if (isident1(c))
        parseIdentifier(st, c);
    }
  } while (isType(token, TOK_NONE));
  return true;
}

/*----------------------------------------------------------------------------
*-	Scanning support functions
----------------------------------------------------------------------------*/

static statementInfo *CurrentStatement = NULL;

static statementInfo *newStatement(statementInfo *const  parent )
{
  statementInfo *const st = new statementInfo;
  unsigned int i;

  for (i = 0  ;  i < (unsigned int)NumTokens  ;  ++i)
    st->token[i] = newToken();

  st->context  = newToken();
  initStatement(st, parent);
  CurrentStatement = st;

  return st;
}

static void deleteStatement()
{
  statementInfo *const st = CurrentStatement;
  statementInfo *const parent = st->parent;
  unsigned int i;

  for (i = 0  ;  i < (unsigned int)NumTokens  ;  ++i)
    delete st->token[i];

  delete st->context;
  delete st;
  CurrentStatement = parent;
}

static void deleteAllStatements()
{
  while (CurrentStatement != NULL)
    deleteStatement();
}

static bool isContextCandidate(const statementInfo *const  st )
{
  const tokenInfo *const prev = prevToken(st, 1);
  bool isCandidate = false;

  if (isType(prev, TOK_KEYWORD)) switch (prev->keyword)
  {
  case KEYWORD_CLASS:
  case KEYWORD_ENUM:
  case KEYWORD_INTERFACE:
  case KEYWORD_NAMESPACE:
  case KEYWORD_STRUCT:
  case KEYWORD_UNION:
    isCandidate = true;
    break;

  default:
    break;
  }
  return isCandidate;
}

static bool isStatementEnd(const statementInfo *const  st )
{
  const tokenInfo *const token = activeToken(st);
  if (isType(token, TOK_SEMICOLON))
    return true;
  else if (isType(token, TOK_BRACE_CLOSE)  &&
    (st->declaration == DECL_FUNCTION  || st->declaration == DECL_NAMESPACE))
    return true;
  else
    return false;
}

static void checkStatementEnd(statementInfo *const  st )
{
  const tokenInfo *const token = activeToken(st);
  if (isType(token, TOK_COMMA))
    reinitStatement(st, true);
  else if (isStatementEnd(st))
    reinitStatement(st, false);
  else
    advanceToken(st);
}

static void nest (statementInfo *const  st,const unsigned int  nestLevel )
{
  switch (st->declaration)
  {
  case DECL_CLASS:
  case DECL_ENUM:
  case DECL_INTERFACE:
  case DECL_NAMESPACE:
  case DECL_STRUCT:
  case DECL_UNION:
    createTags(nestLevel, st);
    break;

  default:
    skipToMatch("{}");
    break;
  }
  advanceToken(st);
  setToken(st, TOK_BRACE_CLOSE);
}

static void tagCheck(statementInfo *const  st )
{
  const tokenInfo *const token = activeToken(st);
  const tokenInfo *const prev  = prevToken(st, 1);
  const tokenInfo *const prev2 = prevToken(st, 2);

  switch (token->type)
  {
  case TOK_NAME:
    if (insideEnumBody(st))
      qualifyEnumeratorTag(st, token);
    break;

  case TOK_BRACE_OPEN:
    if (isType(prev, TOK_ARGS))
    {
      if (st->haveQualifyingName)
      {
        st->declaration = DECL_FUNCTION;
        qualifyBlockTag(st, prev2);
        copyToken(st->context, prev2);
      }
    }
    else
      qualifyBlockTag(st, st->context);
    break;

  case TOK_SEMICOLON:
  case TOK_COMMA:
    if (insideEnumBody(st))
      ;
    else if (isType(prev, TOK_NAME))
      qualifyVariableTag(st, prev);
    else if (isType(prev, TOK_ARGS)  &&  isType(prev2, TOK_NAME))
    {
      if (st->isPointer)
        qualifyVariableTag(st, prev2);
      else
        qualifyFunctionDeclTag(st, prev2);
    }
    break;

  default: break;
  }
}

/*  Parses the current file and decides whether to write out and tags that
*  are discovered.
*/
static void createTags(const unsigned int  nestLevel,statementInfo *const  parent )
{
  statementInfo *const st = newStatement(parent);

  while (true)
  {
    tokenInfo *token;

    if(!nextToken(st))
      break;
    token = activeToken(st);
    if (isType(token, TOK_BRACE_CLOSE))
    {
      if (nestLevel > 0)
        break;
      else
      {
        AfxTrace("%s(%ld): unexpected closing brace",getFileName(), getFileLine());
        throw(ExceptionBraceFormattingError);
      }
    }
    else
    {
      if (isType(token, TOK_NAME) && isContextCandidate(st))
      {
        copyToken(st->context, token);
        st->haveQualifyingName = false;
      }
      else
        tagCheck(st);

      if (isType(token, TOK_BRACE_OPEN))
        nest(st, nestLevel + 1);

      checkStatementEnd(st);
    }
  }
  deleteStatement();
}

void buildCKeywordHash()
{
  const size_t count = sizeof(KeywordTable) / sizeof(KeywordTable[0]);
  size_t i;

  for (i = 0  ;  i < count  ;  ++i)
  {
    const keywordDesc *p = &KeywordTable[i];

    if (p->isValid[0])
      addKeyword(p->name, LANG_DSP, (int)p->id);
    if (p->isValid[1])
      addKeyword(p->name, LANG_CPP, (int)p->id);
  }
}

void createCTags()
{
  cppInit();
  try
  {
    createTags(0, NULL);
  }
  catch(exception_t&)
  {
    deleteAllStatements();
  }
  cppTerminate();
}
